@@ -1,38 +1,162 @@
-# from sandpyper.outils import cross_ref, getListOfFiles, getDate, getLoc, getCrs_from_raster_path
-# from sandpyper.profile import extract_from_folder, get_profiles
-#
-# import rasterio as ras
-# import rasterio.mask as rasmask
-#
-# import os
-# import glob
-#
-# import pandas as pd
-# import re
-# import geopandas as gpd
-# import numpy as np
-# import matplotlib.pyplot as plt
-# import seaborn as sb
-# from shapely.geometry import MultiLineString, LineString, Point, Polygon, MultiPolygon, mapping
-# from shapely.ops import split, snap, unary_union
-#
-#
-# import math
-# from math import tan, radians, sqrt
-# from scipy.ndimage import gaussian_filter
-# import scipy.signal as sig
-# from skimage.filters import threshold_multiotsu
-# from sklearn.metrics import mean_squared_error
-#
-#
-# from tqdm import tqdm
-#
-# import richdem as rd
-# import itertools as it
-# from itertools import groupby
-# from operator import itemgetter
-# import datetime as dt
+from sandpyper.outils import cross_ref, getListOfFiles, getDate, getLoc, getCrs_from_raster_path
+from sandpyper.profile import extract_from_folder, get_profiles
 
+import rasterio as ras
+import rasterio.mask as rasmask
+
+import os
+import glob
+
+import pandas as pd
+import re
+import geopandas as gpd
+import numpy as np
+import matplotlib.pyplot as plt
+import seaborn as sb
+from shapely.geometry import MultiLineString, LineString, Point, Polygon, MultiPolygon, mapping
+from shapely.ops import split, snap, unary_union
+
+
+import math
+from math import tan, radians, sqrt
+from scipy.ndimage import gaussian_filter
+import scipy.signal as sig
+from skimage.filters import threshold_multiotsu
+from sklearn.metrics import mean_squared_error
+
+
+from tqdm import tqdm
+
+import richdem as rd
+import itertools as it
+from itertools import groupby
+from operator import itemgetter
+import datetime as dt
+
+
+
+
+def images_to_dirs(images_folder,target_folder,op=None):
+    """Create one folder per image named as the image. Optionally, move image into it.
+
+    Args:
+        images_folder (str): path to the directory where the images are stored.
+        target_folder (str): target path where create subfolders named as the images.
+        move_images (bool): True, to move the images into the newly created folders.
+    Returns:
+        Create folders (optionally, containing the images).
+
+    """
+
+    images=os.listdir(images_folder)
+    original_images_paths=[os.path.join(images_folder,image) for image in images]
+    ids=[os.path.splitext(image)[0] for image in images]
+
+    if op != None:
+        target_images_paths=[os.path.join(target_folder,id_in) for id_in in ids]
+    # create ID-folders in target_parent_folder, named as image names
+
+    starting_wd=os.getcwd() # store starting working directory
+
+    os.chdir(target_folder) # change working dir
+
+
+    if op == None:
+        for id_i in ids:
+            if not os.path.exists(os.path.join(target_folder,id_i)):
+                os.mkdir(id_i)
+            else:
+                print(f"{id_i} exists already. Skipping this one.")
+
+    elif op != None:
+
+        for id_i,source,destination in zip(ids,original_images_paths,target_images_paths):
+            if not os.path.exists(os.path.join(target_folder,id_i)):
+                os.mkdir(id_i)
+                if op == 'move':
+                    move(source,destination)
+                elif op == 'copy':
+                    copy(source,destination)
+                else:
+                    raise ValueError("op parameter not set. 'move' to move images into newly created folders, 'copy' to copy them or None to only create folders.")
+
+            else:
+                print(f"{id_i} exists already. {op} only.")
+                if op == 'move':
+                    move(source,destination)
+                elif op == 'copy':
+                    copy(source,destination)
+    else:
+        raise ValueError("op parameter not set. 'move' to move images into newly created folders, 'copy' to copy them or None to only create folders.")
+
+    os.chdir(starting_wd) # returning to starting working dir
+
+    print(f"Succesfully created {len(ids)} ID-folders in {target_folder} .")
+
+
+
+def s2_to_rgb(imin, scaler_range=(0,255), re_size=False, dtype=False):
+    scaler=MinMaxScaler(scaler_range)
+
+    imo=ras.open(imin, "r")
+
+    if dtype!=False:
+        im=imo.read(out_dtype=dtype)
+    else:
+        im=imo.read()
+
+    if isinstance(re_size,(tuple)):
+        if imo.count > 1:
+            im=resize(im, (re_size[0], re_size[1], re_size[2]),
+                          mode='constant') # resize image
+        else:
+            im=resize(im, (re_size[1], re_size[2], 1),
+              mode='constant') # resize image
+    else:
+        pass
+
+    if imo.count > 1:
+        im_rgb=np.stack((im[0], im[1], im[2]), axis=-1)
+        rgb_array_1=b = im_rgb.reshape(-1, 1)
+        scaled_1=scaler.fit_transform(rgb_array_1).astype(int)
+        img_array_rgb= scaled_1.reshape(im_rgb.shape)
+    else:
+        img_array_rgb=im
+
+    return img_array_rgb
+
+
+def shoreline_from_prediction(prediction, z, shapely_affine, min_vertices=2, shape=(64,64)):
+    """
+    Extract a georeferenced subpixel shoreline from an array using Marching squares and store it in a GeoDataframe.
+    Credits: adapted from Dr. Robbie Bishop-Taylor functions in Digital Earth Australia scripts, available at:
+    https://github.com/GeoscienceAustralia/dea-notebooks/blob/develop/Scripts/dea_coastaltools.py
+
+    Args:
+        prediction (array): The 2D array returned by the DL model.
+
+        z (float,int): The threshold to use to divide water and no-water.
+
+        shapely_affine (Affine object): Shapely Affine object of the tile the prediction has been performed from.
+
+        min_vertices (int): Minimum number of vertices to retain a shoreline segment (default=2).
+
+        shape (tuple): Shape of the tiles (default= (64,64), minimum requirement for Unet).
+
+    Returns:
+        Grid : A GeoDataFrame storing polygon grids, with IDs and geometry columns.
+    """
+    # get shoreline
+    shore_arr=contours_to_multiline(prediction.reshape(shape),z, min_vertices=min_vertices)
+
+    # create geoseries and geodataframe
+    shore_arr_geoseries=gpd.GeoSeries(shore_arr, name="geometry")
+    contours_gdf=gpd.GeoDataFrame(shore_arr_geoseries, geometry="geometry")
+
+    # georeference line using tile geotransform
+    contours_gdf['geometry'] = contours_gdf.affine_transform(shapely_affine)
+
+    return contours_gdf
 
 def grid_from_pts(pts_gdf, width, height, crs):
     """
@@ -243,172 +367,6 @@ def dissolve_shores(gdf_shores, field='date'):
     return dissolved
 
 
-def tiles_from_grid(
-        grid,
-        img_path,
-        output_path,
-        list_loc_codes,
-        mode='rgb',
-        geotransform=True,
-        sel_bands=None,
-        driver="PNG"):
-    """
-    Returns a dataframe with location, raw_date, filenames (paths) or geopackage index and CRS of each raster and its associated vector files.
-    If the directory containing the vector files has ony one file, it is assumed that this file stores vectors
-    with location and raw_date columns.
-
-    Args:
-        grid (GeoDataFrame): GeoDataFrame of the grid of only tiles containing the line. Output of grid_from_shore function.
-        img_path (str): Path of the directory containing the geotiffs datasets (.tiff or .tif).
-        output_path (str): Path of the directory where to save the images tiles.
-        list_loc_codes (list): list of strings containing location codes.
-        mode (str,'rgb','mask','multi','custom'): 'rgb', the output images are 3-channels RGB tiles. 'mask', 1-band output tiles.
-        'multi', multibands output tiles (with all input bands).
-        'custom', use selected band indices (with sel_bands parameter) to only extract those bands from input multiband images
-        (NOTE: in 'custom' mode, output tile bands indices are reindexd, so do not corresponds with the original band indices, but restart from 1).
-        geotransform (bool or 'only'): If True, save tiles and also return a dictionary with the geotransform of each grid.
-        If False, save tiles without geotransform dictionary. If "only", do not save tiles but return the geotransform dictionary only.
-        sel_bands (list): list of integers (minimum is 1, not zero-indexed) corresponding to the bands to be used to create the tiles. Only used with mode='custom'.
-        Default is None.
-        driver (str, "GTiff" or "PNG"): tiles image file type. Default is "PNG".
-
-    Returns:
-        Saves tiles to the specified output folder and optionally return the tiles geotransform dictionary.
-    """
-
-    if geotransform == 'only':
-        geo_dict = dict()
-        export_image = False
-        export_geo_dict = True
-
-    elif bool(geotransform):
-        geo_dict = dict()
-        export_image = True
-        export_geo_dict = True
-    elif bool(geotransform) == False:
-        export_image = True
-        export_geo_dict = False
-
-    with ras.open(img_path, 'r') as dataset:
-
-        if mode == 'rgb':
-
-            count = 3
-            source_idx = [1, 2, 3]  # the band indices of the source image
-            out_idx = source_idx
-            sel_bands = None
-
-            height_idx = 1
-            width_idx = 2
-
-        elif mode == 'multi':
-
-            if driver == "PNG":
-                print("NOTE: PNG format doesn't support multibands. Returning GeoTiffs instead.")
-                driver = "GTiff"
-            else:
-                pass
-
-            sel_bands = None
-            count = dataset.count
-            source_idx = list(dataset.indexes)
-            out_idx = None
-            height_idx = 1
-            width_idx = 2
-
-        elif mode == 'mask':
-
-            count = 1
-            source_idx = 1
-            out_idx = 1
-            height_idx = 0
-            width_idx = 1
-
-        elif mode == 'custom':
-            if len(sel_bands) > 3 and driver == "PNG":
-                print("NOTE: More than 3 bands selected for the creation of PNG images. PNG format doesn't support multibands. Returning GeoTiffs instead.")
-                driver = "GTiff"
-            else:
-                pass
-
-            source_idx = sel_bands
-            out_idx = None
-            count = len(sel_bands)
-            height_idx = 1
-            width_idx = 2
-
-        for i in range(grid.shape[0]):
-            mask = grid.iloc[[i]]
-            geom = Polygon(   # create the square polygons to clip the raster with
-
-                ((mask.iloc[0]['ulx'], mask.iloc[0]['uly']),
-                 (mask.iloc[0]['urx'], mask.iloc[0]['ury']),
-                    (mask.iloc[0]['lrx'], mask.iloc[0]['lry']),
-                    (mask.iloc[0]['llx'], mask.iloc[0]['lly']),
-                    (mask.iloc[0]['ulx'], mask.iloc[0]['uly']))
-
-            )
-
-            try:
-                out_image, out_transform = rasmask.mask(
-                    dataset, [mapping(geom)], crop=True, filled=False, indexes=source_idx)
-                out_meta = dataset.meta
-
-                # format name: LocationCode_GridId_Date.tif
-                if driver == "PNG":
-                    ext = 'png'
-                elif driver == 'GTiff':
-                    ext = 'tif'
-                else:
-                    raise NameError("Driver must be either 'PNG' or 'GTiff'.")
-
-                if driver == "PNG":
-                    ext = 'png'
-                    out_meta.update({
-                        'dtype': ras.uint16})
-                else:
-                    ext = 'tif'
-
-                out_meta.update({"driver": driver,
-                                 "height": out_image.shape[height_idx],
-                                 "width": out_image.shape[width_idx],
-                                 "transform": out_transform,
-                                 "count": count})
-
-                tile_name = f"{getLoc(img, list_loc_codes)}_{mask.iloc[0]['grid_id']}_{getDate(img)}_{mode}"
-                savetxt = os.path.join(output_path, f"{tile_name}.{ext}")
-
-                if bool(export_image):
-                    with ras.open(savetxt, "w", **out_meta) as dest:
-                        if driver == "PNG":
-                            dest.write(out_image.astype(ras.uint16), indexes=out_idx)
-                        else:
-                            dest.write(out_image, indexes=out_idx)
-                    print(f"Succesfully saved tile: {tile_name} .")
-
-                    if bool(export_geo_dict):
-                        geo_dict.update({tile_name: out_transform})
-                    else:
-                        pass
-
-                else:
-                    pass
-
-                if bool(export_geo_dict):
-                    geo_dict.update({tile_name: out_transform})
-                else:
-                    pass
-
-            except BaseException:
-                print(
-                    f"Error with tile {mask.iloc[0]['grid_id']}. Might not overlap iamge. Skipping.")
-
-    if bool(export_geo_dict):
-        return geo_dict
-    else:
-        pass
-
-
 def check_overlay(line_geometry, img_path):
     """ Evaluate wether a line intesects the extent of a raster.
         Returns True if a valid intersection is found or False if not. In case of MultiLine features,
@@ -1623,3 +1581,435 @@ def save_slope_corr_files(
 
         print("Ready to correct shorelines.")
     return gdf
+
+
+def s2_to_rgb(imin, scaler_range=(0,255), re_size=False, dtype=False):
+    scaler=MinMaxScaler(scaler_range)
+
+    imo=ras.open(imin, "r")
+    if dtype!=False:
+        im=imo.read(out_dtype=dtype)
+    else:
+        im=imo.read()
+
+    if isinstance(re_size,(tuple)):
+        im=resize(im, (re_size[0], re_size[1], re_size[2]),
+                      mode='constant', preserve_range=True) # resize image
+    else:
+        pass
+
+    if imo.count > 1:
+        im_rgb=np.stack((im[0], im[1], im[2]), axis=-1)
+        rgb_array_1=b = im_rgb.reshape(-1, 1)
+        scaled_1=scaler.fit_transform(rgb_array_1).astype(int)
+        img_array_rgb= scaled_1.reshape(im_rgb.shape)
+    else:
+        img_array_rgb=im
+
+    if isinstance(re_size,(tuple)):
+        img_array_rgb=resize(im, (re_size[0], re_size[1], re_size[2]), mode='constant', preserve_range=True) # resize image
+
+    return img_array_rgb
+
+
+def partial_tile_padding(dataset,
+                geom,
+                expected_shape,
+                crs,
+                tile_name,
+                output_path,
+                nodata,
+                source_idx,
+                height_idx,
+                width_idx,
+                out_idx,
+                count,
+                driver,
+                ):
+
+
+
+   ## This will be the tile name
+
+    ## Create a padded tile from the mask coordinates
+
+    px_size=dataset.transform[0]
+    ulx=geom.bounds[0]
+    ury=geom.bounds[3]
+
+    ulx_px_coords=floor(int(ulx)/10)*10
+    ury_px_coords=ceil(int(ury)/10)*10
+
+    pad_transform = from_origin(ulx_px_coords,
+                            ury_px_coords,
+                            px_size,px_size)
+
+    pad_array=ras.features.rasterize([mapping(geom)], out_shape=expected_shape,
+                                    all_touched=False)
+    print(f"pad_array:{pad_array.shape}")
+
+    #pad_array_3=np.array([pad_array, pad_array, pad_array], ras.int16)
+    pad_array_3=np.dstack([pad_array]*count).astype(ras.int16).reshape((count,expected_shape[0],expected_shape[1]))
+
+    print(f"pad_array_3:{pad_array_3.shape}")
+    print(f"height_idx:{height_idx}")
+    print(f"width_idx:{width_idx}")
+
+
+    with MemoryFile(filename="padded") as memfile_padded:
+        with memfile_padded.open(driver=driver,
+            height=pad_array_3.shape[1],
+            width=pad_array_3.shape[2],
+            count=count,
+            dtype=pad_array_3.dtype,
+            crs=crs,
+            transform=pad_transform) as mem_dataset_padded:
+
+            mem_dataset_padded.write(pad_array_3, indexes=list(np.arange(1,count+1)))
+
+
+            # Create a the temporary in-memory partial tile
+
+            out_image, out_transform = rasmask.mask(dataset=dataset,
+                                                    shapes=[mapping(geom)],
+                                                    invert=False,
+                                                    crop=True,
+                                                    filled=False,
+                                                    nodata=nodata,
+                                                    indexes=source_idx)
+
+
+            out_meta = dataset.meta
+            print(f"out_image:{out_image.shape}")
+
+            # format name: LocationCode_GridId_Date.tif
+            if driver=="PNG":
+                ext='png'
+            elif driver=='GTiff':
+                ext='tif'
+            else:
+                raise NameError("Driver must be either 'PNG' or 'GTiff'.")
+
+            if driver=="PNG":
+                ext='png'
+                out_meta.update({
+                'dtype':ras.uint16})
+            else:
+                ext='tif'
+
+
+            out_meta.update({
+             "height": out_image.shape[height_idx],
+             "width": out_image.shape[width_idx],
+             "transform": out_transform,
+            "count":count})
+
+
+            with MemoryFile() as memfile:
+                with memfile.open(**out_meta) as mem_dataset:
+
+                    if driver=="PNG":
+                        mem_dataset.write(out_image.astype(ras.uint16), indexes=out_idx)
+                    else:
+                        mem_dataset.write(out_image,indexes=out_idx)
+
+
+                    # mosaic in-memory rasters and write to disk
+                    two_rasters=[mem_dataset,mem_dataset_padded]
+
+                    mosaic, out_trans = merge(datasets=two_rasters,
+                                              indexes=list(np.arange(1,count+1)),
+                                              nodata=nodata,
+                                              method="max")
+
+                    print(f"mosaic:{mosaic.shape}")
+
+                    savetxt=os.path.join(output_path,tile_name)
+
+
+                    with ras.open(
+                        savetxt,
+                        'w',
+                        driver=driver,
+                        height=mosaic.shape[1],
+                        width=mosaic.shape[2],
+                        count=count,
+                        dtype=mosaic.dtype,
+                        crs=crs,
+                        transform=out_trans) as final_mosaic_tile:
+
+                        final_mosaic_tile.write(mosaic)
+
+                    print(f"Succesfully saved partial tile in-memory: {tile_name} .")
+                    if geotransform==True:
+
+                            geot_series=pd.Series({'tile_code':f"{tile_code}",
+                                                   'geotransform':out_trans})
+
+                            print("Tile geotransform returned.")
+                            return geot_series
+
+
+def tile_to_disk(dataset,
+                geom,
+                crs,
+                tile_name,
+                output_path,
+                nodata,
+                source_idx,
+                height_idx,
+                width_idx,
+                out_idx,
+                count,
+                driver,
+                ):
+
+    try:
+        out_image, out_transform = rasmask.mask(dataset=dataset,
+                                                shapes=[mapping(geom)],
+                                                invert=False,
+                                                crop=True,
+                                                filled=False,
+                                                nodata=nodata,
+                                                indexes=source_idx)
+    except:
+        out_image, out_transform = rasmask.mask(dataset=dataset,
+                                                shapes=geom,
+                                                invert=False,
+                                                crop=True,
+                                                filled=False,
+                                                nodata=nodata,
+                                                indexes=source_idx)
+    out_meta = dataset.meta
+
+
+    if driver=="PNG":
+        out_meta.update({
+        'dtype':ras.uint16})
+
+
+    out_meta.update(
+                    {"driver": driver,
+                     "height": out_image.shape[height_idx],
+                     "width": out_image.shape[width_idx],
+                     "transform": out_transform,
+                     "count":count,
+                    "crs":crs}
+                   )
+
+    savetxt=os.path.join(output_path,tile_name)
+
+
+    with ras.open(savetxt, "w", **out_meta) as dest:
+        if driver=="PNG":
+            dest.write(out_image.astype(ras.uint16), indexes=out_idx)
+        else:
+            dest.write(out_image,indexes=out_idx)
+
+    print(f"Succesfully saved tile: {tile_name} .")
+    if geotransform==True:
+
+            geot_series=pd.Series({'tile_code':f"{tile_code}",
+                                   'geotransform':out_transform})
+
+            print("Tile geotransform returned.")
+            return geot_series
+
+
+def tiles_from_grid (grid,img_path,
+                     output_path,
+                     list_loc_codes,crs_dict_string,
+                     mode='rgb',
+                     sel_bands=None,
+                     driver="PNG"):
+    """
+    Returns a dataframe with location, raw_date, filenames (paths) or geopackage index and CRS of each raster and its associated vector files.
+    If the directory containing the vector files has ony one file, it is assumed that this file stores vectors
+    with location and raw_date columns.
+
+    Args:
+        grid (GeoDataFrame): GeoDataFrame of the grid of only tiles containing the line. Output of grid_from_shore function.
+        img_path (str): Path of the directory containing the geotiffs datasets (.tiff or .tif).
+        output_path (str): Path of the directory where to save the images tiles.
+        list_loc_codes (list): list of strings containing location codes.
+        mode (str,'rgb','mask','multi','custom'): 'rgb', the output images are 3-channels RGB tiles. 'mask', 1-band output tiles.
+        'multi', multibands output tiles (with all input bands).
+        'custom', use selected band indices (with sel_bands parameter) to only extract those bands from input multiband images
+        (NOTE: in 'custom' mode, output tile bands indices are reindexd, so do not corresponds with the original band indices, but restart from 1).
+        geotransform (bool or 'only'): If True, save tiles and also return a dictionary with the geotransform of each grid.
+        If False, save tiles without geotransform dictionary. If "only", do not save tiles but return the geotransform dictionary only.
+        sel_bands (list): list of integers (minimum is 1, not zero-indexed) corresponding to the bands to be used to create the tiles. Only used with mode='custom'.
+        Default is None.
+        driver (str, "GTiff" or "PNG"): tiles image file type. Default is "PNG".
+
+    Returns:
+        Saves tiles to the specified output folder and optionally return the tiles geotransform dictionary.
+    """
+
+    loc=getLoc(img_path,list_loc_codes)
+    crs=crs_dict_string[loc]
+
+    if driver=="PNG":
+        ext='png'
+    elif driver=='GTiff':
+        ext='tif'
+    else:
+        raise NameError("Driver must be either 'PNG' or 'GTiff'.")
+
+    with ras.open(img_path,'r') as dataset:
+
+        if mode=='rgb':
+
+            count=3
+            source_idx=[1,2,3] # the band indices of the source image
+            out_idx=source_idx
+            sel_bands=None
+
+            height_idx=1
+            width_idx=2
+
+        elif mode=='multi':
+
+            if driver=="PNG":
+                print("NOTE: PNG format doesn't support multibands. Returning GeoTiffs instead.")
+                driver="GTiff"
+            else:
+                pass
+
+            sel_bands=None
+            count=dataset.count
+            source_idx=list(dataset.indexes)
+            out_idx=None
+            height_idx=1
+            width_idx=2
+
+        elif mode=='mask':
+
+            count=1
+            source_idx=1
+            out_idx=1
+            height_idx=0
+            width_idx=1
+
+        elif mode=='custom':
+            if len(sel_bands)>3 and driver=="PNG":
+                print("NOTE: More than 3 bands selected for the creation of PNG images. PNG format doesn't support multibands. Returning GeoTiffs instead.")
+                driver="GTiff"
+            else:
+                pass
+
+            source_idx=sel_bands
+            out_idx=None
+            count=len(sel_bands)
+            height_idx=1
+            width_idx=2
+
+        # creates gereferenced bounding box of the image
+        geom_im = gpd.GeoSeries(box(*dataset.bounds), crs=crs_dict_string[loc])
+
+        # evaluates which tiles are fully within the raster bounds
+        fully_contains=[geom_im.geometry.contains(mask_geom)[0] for mask_geom in grid.geometry]
+
+        # get the expected shape of fully contained tiles
+        full_in_geom=grid[fully_contains].iloc[[0]]["geometry"]
+        geom_wdw=geometry_window(dataset,full_in_geom)
+        expected_shape=(geom_wdw.height,geom_wdw.width)
+
+        print(f"Expected shape{expected_shape}")
+
+        for i, row in grid.iterrows():
+
+            tile_name=f"{loc}_{row.grid_id}_{getDate(img_path)}_{mode}.{ext}"
+
+            geom=Polygon(   # create the square polygons to clip the raster with
+
+               ((row.ulx,row.uly),
+               (row.urx,row.ury),
+               (row.lrx,row.lry),
+               (row.llx,row.lly),
+                (row.ulx,row.uly))
+
+                    )
+
+
+            # get the future shape of the tile which is about to get created
+            geom_wdw=geometry_window(im,[mapping(geom)])
+            tile_shape=(geom_wdw.height,geom_wdw.width)
+
+
+            if tile_shape == expected_shape:
+
+                tile_to_disk(dataset=dataset,
+                             geom=geom,
+                             crs=crs,
+                             tile_name=tile_name,
+                             output_path=output_path,
+                             nodata=0,
+                             source_idx=source_idx,
+                             height_idx=height_idx,
+                             width_idx=width_idx,
+                             out_idx=out_idx,
+                             count=count,
+                             driver=driver
+                                )
+
+            else:
+
+                partial_tile_padding(dataset=dataset,
+                                     expected_shape=expected_shape,
+                                     crs=crs,
+                                     geom=geom,
+                                     tile_name=tile_name,
+                                     output_path=output_path,
+                                     nodata=0,
+                                     source_idx=source_idx,
+                                     height_idx=height_idx,
+                                     width_idx=width_idx,
+                                     out_idx=out_idx,
+                                     count=count,
+                                     driver=driver
+                                    )
+def arr2geotiff (array, transform, location, shape=(64,64,1),driver="GTiff", dtype=np.float32, save=None):
+    with MemoryFile() as memfile:
+        mem_dataset= memfile.open(driver="GTiff",
+                           height = array.shape[0],
+                           width = array.shape[1],
+                           count=array.shape[2],
+                           dtype=dtype,
+                          transform=transform,
+                           crs=crs_dict_string[location])
+
+        mem_dataset.write(array.reshape((shape[0],shape[1])),indexes=shape[2])
+
+
+        if save != None:
+            with ras.open(save, "w",
+                         driver=driver,
+                        height=array.shape[0],
+                        width=array.shape[1],
+                        count=array.shape[2],
+                        dtype=dtype,
+                        transform=transform,
+                          crs=crs_dict_string[location]
+                         ) as dest:
+                if driver=="PNG":
+                    dest.write(mem_dataset.astype(ras.uint16), indexes=1)
+                else:
+                    dest.write(array.reshape((shape[0],shape[1])),indexes=shape[2])
+
+
+        return mem_dataset
+
+def shoreline_from_prediction(prediction, z, geotransform, min_vertices=2, shape=(64,64)):
+
+    # get shoreline
+    shore_arr=contours_to_multiline(prediction.reshape(shape),z, min_vertices=min_vertices)
+
+    # create geoseries and geodataframe
+    shore_arr_geoseries=gpd.GeoSeries(shore_arr, name="geometry")
+    contours_gdf=gpd.GeoDataFrame(shore_arr_geoseries, geometry="geometry")
+
+    # georeference line using tile geotransform
+    contours_gdf['geometry'] = contours_gdf.affine_transform(shapely_affine)
+
+    return contours_gdf
